"""Helpers for tracking collection counts.

There are two types of dictionary used in this module: card_row, and
print_counts. These dictionaries are of the following form:

print_counts:
    used for tracking aggregate counts of cards in a collection

    signature: Dict[models.CardPrinting, Dict[CountTypes, int]]

card_row:
    used for representing a single card and any counts for it

    form: {
        'id': str, 'set': str, 'name': str, 'number': str,
        'multiverseid': str/int, CountTypes.name: str/int,
        [CountTypes.name: str/int, [...]]
    }

The helper methods in this modules are primarily concerned with converting
iterables of card_row dicts into singular print_counts dicts.
"""

import collections
import enum
from typing import Mapping

NAME_SUBSTITUTIONS = [
    ('Ae', 'Æ'),
    ('Jo', 'Jö'),
]
NAME_SUBSTITUTIONS += [(right, left) for (left, right) in NAME_SUBSTITUTIONS]


class Error(Exception):
    """Base exception for this module."""


class UnknownPrintingError(Error):
    """Raised when a printing cannot be found in count conversions."""


class CountTypes(enum.Enum):
    """Enum for possible card printing types (normal, foil)."""
    copies = 'copies'
    foils = 'foils'


def new_print_counts() -> Mapping[str, Mapping[CountTypes, int]]:
    """Get an appropriate defaultdict set up for use ase print counts."""
    return collections.defaultdict(collections.Counter)


def find_printing(cdb, set_code, name, set_number, multiverseid, strict=True):
    """Attempt to find a CardPrinting from given parameters."""
    print('Searching => Set: %s; Name: %s; Number: %s, Multiverse ID: %s' % (
        set_code, name, set_number, multiverseid))
    name = name or ''
    names = [name]
    for sub_from, sub_to in NAME_SUBSTITUTIONS:
        if sub_from in name:
            names.append(name.replace(sub_from, sub_to))
    snnm_keys = []
    for name_var in names:
        snnm_keys.extend([
            (set_code, name_var, set_number, multiverseid),
            (set_code, name_var, None, multiverseid),
            (set_code, name_var, set_number, None),
            (set_code, name_var, None, None),
        ])
    printing = None
    for snnm_key in snnm_keys:
        found_printings = cdb.set_name_num_mv_to_printings.get(snnm_key, [])
        if len(found_printings) == 1:
            printing = found_printings[0]
            break
        elif found_printings and not strict:
            found_printings = sorted(found_printings, key=lambda p: p.id_)
            printing = found_printings[0]
            break
    if printing is not None:
        print('Found => Set: %s; Name: %s; Number: %s; Multiverse ID: %s' % (
            printing.set_code, printing.card_name, printing.set_number,
            printing.multiverseid))
    return printing


def coerce_card_row(card_count):
    """Given a card_row dict, coerce types to match desired input."""
    int_keys = ['multiverseid'] + [ct.name for ct in CountTypes]
    for key in int_keys:
        try:
            card_count[key] = int(card_count[key])
        except (KeyError, TypeError, ValueError):
            pass
    return card_count


def aggregate_print_counts(cdb, card_rows, strict):
    """Given a card database Iterable[card_row], return print_counts"""
    print_counts = new_print_counts()
    for card_row in card_rows:
        if not any(card_row.get(ct.name) for ct in CountTypes):
            continue
        card_row = coerce_card_row(card_row)
        printing_id = card_row.get('id')
        printing = cdb.id_to_printing.get(printing_id)
        if printing is None:
            print('printing not found by id, searching')
            printing = find_printing(
                cdb=cdb,
                set_code=card_row.get('set'),
                name=card_row.get('name'),
                set_number=card_row.get('number'),
                multiverseid=card_row.get('multiverseid'),
                strict=strict)
        if printing is None:
            raise UnknownPrintingError(
                'Could not match id to known printing from counts: %r' %
                card_row)
        for count_type in CountTypes:
            ct_name = count_type.name
            count = card_row.get(ct_name)
            if count:
                print_counts[printing.id_][count_type] += count
    return print_counts


def merge_print_counts(*print_counts_args):
    """Merge two sets of print_counts."""
    merged_counts = new_print_counts()
    for print_counts in print_counts_args:
        for print_id, counts in print_counts.items():
            merged_counts[print_id].update(counts)
    return merged_counts


def diff_print_counts(left, right):
    """Subtract right print counts from left print counts."""
    print_counts = new_print_counts()
    for print_id in left.keys() | right.keys():
        left_counts = left.get(print_id, {})
        right_counts = right.get(print_id, {})
        for key in left_counts.keys() | right_counts.keys():
            value = left_counts.get(key, 0) - right_counts.get(key, 0)
            if value:
                print_counts[print_id][key] = value
    return print_counts



"""CSV serializer."""

import csv

from mtg_ssm.mtg import counts
from mtg_ssm.serialization import interface

CSV_HEADER = [
    'set',
    'name',
    'number',
    'multiverseid',
    'id',
] + [ct.name for ct in counts.CountTypes]


def row_for_printing(printing, printing_counts):
    """Given a CardPrinting and counts, return a csv row."""
    csv_row = {
        'set': printing.set_code,
        'name': printing.card_name,
        'number': printing.set_number,
        'multiverseid': printing.multiverseid,
        'id': printing.id_,
    }
    for counttype, count in printing_counts.items():
        if count:
            csv_row[counttype.name] = count
    return csv_row


def rows_for_printings(cdb, print_counts, verbose):
    """Generator that yields csv rows from a card_db."""
    for card_set in cdb.card_sets:
        for printing in card_set.printings:
            printing_counts = print_counts.get(printing.id_, {})
            if verbose or any(printing_counts):
                yield row_for_printing(printing, printing_counts)


class CsvFullDialect(interface.SerializationDialect):
    """csv collection writing a row for every printing"""
    extension = 'csv'
    dialect = 'csv'

    verbose = True

    def write(self, filename: str, print_counts) -> None:
        """Write print counts to a file."""
        with open(filename, 'w') as csv_file:
            writer = csv.DictWriter(csv_file, CSV_HEADER)
            writer.writeheader()
            for row in rows_for_printings(
                    self.cdb, print_counts, self.verbose):
                writer.writerow(row)

    def read(self, filename: str):
        """Read print counts from file."""
        with open(filename, 'r') as csv_file:
            return counts.aggregate_print_counts(
                self.cdb, csv.DictReader(csv_file), strict=True)


class CsvTerseDialect(CsvFullDialect):
    """csv collection writing only rows that have counts"""
    dialect = 'terse'

    verbose = False



#!/usr/bin/env python3
"""Script for managing magic card spreadsheets."""

import argparse
import datetime
import os
import shutil
import tempfile

import mtg_ssm

from mtg_ssm import mtgjson
import mtg_ssm.serialization.interface as ser_interface
from mtg_ssm.mtg import card_db
from mtg_ssm.mtg import counts


DEFAULT_DATA_PATH = os.path.expanduser(os.path.join('~', '.mtg_ssm'))


def epilog():
    """Generate the argparse help epilog with dialect descriptions."""
    dialect_docs = 'available dialects:\n'
    ext_dia_desc = ser_interface.SerializationDialect.dialects()
    for extension, dialect, description in ext_dia_desc:
        dialect_docs += '  {ext:<8} {dia:<12} {desc}\n'.format(
            ext=extension, dia=dialect, desc=description)
    return dialect_docs


def get_args(args=None):
    """Parse and return application arguments."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description='Magic collection Spreadsheet Manager',
        epilog=epilog())
    parser.add_argument(
        '--version', action='version', version=mtg_ssm.__version__)

    parser.add_argument(
        '--data-path', default=DEFAULT_DATA_PATH,
        help='Path to mtg_ssm\'s data storage folder. Default={}'.format(
            DEFAULT_DATA_PATH))
    parser.add_argument(
        '--include-online-only', default=False, action='store_true',
        help='Include online only sets (e.g. Masters sets)')

    parser.add_argument(
        '-d', '--dialect', nargs=2, metavar=('EXTENSION', 'DIALECT'),
        action='append', default=[],
        help='Mapping of file extensions to serializer dialects. '
             'May be repeated for multiple different extensions.')

    # Commands
    subparsers = parser.add_subparsers(dest='action', title='actions')
    subparsers.required = True

    create = subparsers.add_parser(
        'create', aliases=['c'],
        help='Create a new, empty collection spreadsheet')
    create.set_defaults(func=create_cmd)
    create.add_argument(
        'collection', help='Filename for the new collection')

    update = subparsers.add_parser(
        'update', aliases=['u'],
        help='Update cards in a collection spreadsheet, preserving counts')
    update.set_defaults(func=update_cmd)
    update.add_argument(
        'collection', help='Filename for the collection to update')

    merge = subparsers.add_parser(
        'merge', aliases=['m'],
        help='Merge one or more collection spreadsheets into another. May '
        'also be used for format conversions.')
    merge.set_defaults(func=merge_cmd)
    merge.add_argument(
        'collection', help='Filename for the target collection')
    merge.add_argument(
        'imports', nargs='+',
        help='Filename(s) for collection(s) to import/merge counts from')

    diff = subparsers.add_parser(
        'diff', aliases=['d'],
        help='Create a collection from the differences between two other '
        'collections')
    diff.set_defaults(func=diff_cmd)
    diff.add_argument(
        'left',
        help='Filename for first collection to diff (positive counts)')
    diff.add_argument(
        'right',
        help='Filename for second collection to diff (negative counts)')
    diff.add_argument(
        'output', help='Filename for result collection of diff')

    parsed_args = parser.parse_args(args=args)
    parsed_args.dialect = dict(parsed_args.dialect)
    return parsed_args


def build_card_db(data_path, include_online_only):
    """Get a card_db with current mtgjson data."""
    try:
        mtgjson.fetch_mtgjson(data_path)
    except mtgjson.DownloadError:
        print('Failed to fetch mtgjson data, attempting to use cached data.')
    print('Reading mtgjson data.')
    mtgjsondata = mtgjson.read_mtgjson(data_path)
    return card_db.CardDb(
        mtgjsondata, include_online_only=include_online_only)


def get_serializer(cdb, dialect_mapping, filename):
    """Retrieve a serializer compatible with a given filename."""
    _, extension = os.path.splitext(filename)
    extension = extension.lstrip('.')
    serialization_class = ser_interface.SerializationDialect.by_extension(
        extension, dialect_mapping)
    return serialization_class(cdb)


def get_backup_name(filename):
    """Given a filename, return a timestamped backup name for the file."""
    basename, extension = os.path.splitext(filename)
    extension = extension.lstrip('.')
    now = datetime.datetime.now()
    return '{basename}.{now:%Y%m%d_%H%M%S}.{extension}'.format(
        basename=basename, now=now, extension=extension)


def write_file(serializer, print_counts, filename):
    """Write print counts to a file, backing up existing target files."""
    if not os.path.exists(filename):
        print('Writing collection to file: ' + filename)
        serializer.write(filename, print_counts)
    else:
        backup_name = get_backup_name(filename)
        with tempfile.NamedTemporaryFile() as temp_coll:
            print('Writing to temporary file.')
            serializer.write(temp_coll.name, print_counts)
            print('Backing up exiting file to: ' + backup_name)
            shutil.copy(filename, backup_name)
            print('Overwriting with new collection: ' + filename)
            shutil.copy(temp_coll.name, filename)


def create_cmd(args):
    """Create a new, empty collection."""
    cdb = build_card_db(args.data_path, args.include_online_only)
    serializer = get_serializer(cdb, args.dialect, args.collection)
    write_file(serializer, {}, args.collection)


def update_cmd(args):
    """Update an existing collection, preserving counts."""
    cdb = build_card_db(args.data_path, args.include_online_only)
    serializer = get_serializer(cdb, args.dialect, args.collection)
    print('Reading counts from ' + args.collection)
    print_counts = serializer.read(args.collection)
    write_file(serializer, print_counts, args.collection)


def merge_cmd(args):
    """Merge counts from one or more inputs into a new/existing collection."""
    cdb = build_card_db(args.data_path, args.include_online_only)
    coll_serializer = get_serializer(cdb, args.dialect, args.collection)
    print_counts = counts.new_print_counts()
    if os.path.exists(args.collection):
        print('Reading counts from ' + args.collection)
        print_counts = coll_serializer.read(args.collection)
    for import_file in args.imports:
        input_serializer = get_serializer(cdb, args.dialect, import_file)
        print('Merging counts from ' + import_file)
        print_counts = counts.merge_print_counts(
            print_counts, input_serializer.read(import_file))
    write_file(coll_serializer, print_counts, args.collection)


def diff_cmd(args):
    """Diff two collections, putting the output in a third."""
    cdb = build_card_db(args.data_path, args.include_online_only)
    left_serializer = get_serializer(cdb, args.dialect, args.left)
    right_serializer = get_serializer(cdb, args.dialect, args.right)
    output_serializer = get_serializer(cdb, args.dialect, args.output)
    print('Diffing counts between %s and %s' % (args.left, args.right))
    print_counts = counts.diff_print_counts(
        left_serializer.read(args.left),
        right_serializer.read(args.right))
    write_file(output_serializer, print_counts, args.output)


def main():
    """Get args and run the appropriate command."""
    args = get_args()
    args.func(args)


if __name__ == '__main__':
    main()
